@@ -1,8 +1,10 @@
+from __future__ import annotations
+from collections import defaultdict
 import re
-from dataclasses import dataclass, asdict
+from dataclasses import dataclass, asdict, field
 from hashlib import md5
 from pathlib import Path
-from typing import Iterable, List, Optional, Union, Set
+from typing import Dict, Iterable, List, Optional, Tuple, Union, Set
 from urllib.parse import urlparse
 
 import fastnumbers
@@ -11,6 +13,7 @@ from slugify import slugify
 
 import sm.misc as M
 from sm.inputs.table import ColumnBasedTable, Column
+from grams.inputs.context import ContentHierarchy
 
 
 @dataclass
@@ -19,10 +22,10 @@ class LinkedTable:
     table: ColumnBasedTable
 
     # context of the table
-    context: "Context"
+    context: Context
 
     # a mapping from (row id, column id) to the list of links attach to that cell (not include header)
-    links: List[List[List["Link"]]]
+    links: List[List[List[Link]]]
 
     @property
     def id(self):
@@ -44,42 +47,55 @@ class LinkedTable:
             return 0
         return len(self.table.columns[0].values)
 
+    def iter_cell_index(self):
+        nrows, ncols = self.table.shape()
+        for ri in range(nrows):
+            for ci in range(ncols):
+                yield ri, ci
+
     def to_dict(self):
         return {
-            "version": "1",
+            "version": "1.2",
             "table": self.table.to_dict(),
-            "context": asdict(self.context),
+            "context": self.context.to_dict(),
             "links": [
-                [[asdict(link) for link in links] for links in rlinks]
+                [[link.to_dict() for link in links] for links in rlinks]
                 for rlinks in self.links
             ],
         }
 
     def get_friendly_fs_id(self):
-        id = self.id
+        return LinkedTable._get_friendly_fs_id(self.id)
+
+    @staticmethod
+    def _get_friendly_fs_id(id: str):
         if id.startswith("http://") or id.startswith("https://"):
             if id.find("dbpedia.org") != -1:
-                id = slugify(
-                    urlparse(id).path.replace("/resource/", "").replace("/", "_")
-                ).replace("-", "_")
-                id += "_" + md5(self.id.encode()).hexdigest()
-            elif id.find("wikipedia.org") != -1:
-                id = slugify(
-                    urlparse(id).path.replace("/wiki/", "").replace("/", "_")
-                ).replace("-", "_")
-                id += "_" + md5(self.id.encode()).hexdigest()
-            else:
-                raise NotImplementedError()
-        else:
-            id = slugify(self.id.replace("/", "_")).replace("-", "_")
+                return (
+                    slugify(
+                        urlparse(id).path.replace("/resource/", "").replace("/", "_")
+                    ).replace("-", "_")
+                    + "_"
+                    + md5(id.encode()).hexdigest()
+                )
 
-        return id
+            if id.find("wikipedia.org") != -1:
+                return (
+                    slugify(
+                        urlparse(id).path.replace("/wiki/", "").replace("/", "_")
+                    ).replace("-", "_")
+                    + "_"
+                    + md5(id.encode()).hexdigest()
+                )
+
+            raise NotImplementedError()
+        return slugify(id.replace("/", "_")).replace("-", "_")
 
     @staticmethod
     def from_dict(odict: dict):
-        assert str(odict.get("version", None)) == "1.1"
+        assert float(odict.get("version", "0.0")) >= 1.1
         tbl = ColumnBasedTable.from_dict(odict["table"])
-        context = Context(**odict["context"])
+        context = Context.from_dict(odict["context"])
         links = [
             [[Link.from_dict(link) for link in links] for links in rlinks]
             for rlinks in odict["links"]
@@ -103,16 +119,10 @@ class LinkedTable:
         first_row_header: bool = True,
         table_id: Optional[str] = None,
         top_k: int = 100,
-    ) -> "LinkedTable":
+    ) -> LinkedTable:
         """Load table from a csv file, and its links from a tsv file (if exist).
 
-        Each row of a link file has the following format: `<row_index>\t<col_index>\t(<link>|(<entity_id>(\t<entity_id>)*))`, where:
-            * `row_index` and `col_index` start from 0
-            * `row_index` does not count the header of the table (i.e., skip the first row of `infile` if it's the header)
-            * `(<link>|(<entity_id>(\t<entity_id>)*))` is either:
-                - `<link>` a json string encoding Link object and can be deserialized using `Link.from_dict` function
-                - or (<entity_id>(\t<entity_id>)*) a list of entity ids joined by `\t` tab character, each entity id can be a wikidata qnode id (e.g., Q414) or a full qnode uri (e.g., "http://www.wikidata.org/entity/Q414"). The first entity is considered as the correct entity of the cell, and the rest are considered as the candidate entities of the cell
-        Note that a pair `<row_index>`, `<col_index>` don't have to be unique.
+        For format of the link file, see `LinkedTable.parse_link_file`
 
         Args:
             infile: csv file
@@ -142,55 +152,78 @@ class LinkedTable:
         for ci, cname in enumerate(headers):
             columns.append(Column(ci, cname, [r[ci] for r in rows]))
         table = ColumnBasedTable(table_id, columns)
-        links = []
-        for ri in range(len(rows)):
-            links.append([[] for _ci in range(len(headers))])
 
         if link_file.exists():
-            rows = []
-            for row in M.deserialize_csv(link_file, delimiter="\t"):
-                ri, ci, ents = int(row[0]), int(row[1]), row[2:]
-                rows.append((ri, ci, ents))
+            links = LinkedTable.parse_link_file(table, link_file, top_k)
+        else:
+            links = []
+            for ri in range(len(rows)):
+                links.append([[] for _ci in range(len(headers))])
 
-            has_only_correct_entity = all(len(ents) == 1 for _, _, ents in rows)
-
-            for ri, ci, ents in rows:
-                if len(ents) == 1 and ents[0].startswith("{"):
-                    link = Link.from_dict(orjson.loads(ents[0]))
-                else:
-                    pents = []
-                    for ent in ents[: top_k + 1]:
-                        if ent.startswith("http"):
-                            assert ent.startswith("http://www.wikidata.org/entity/")
-                            qnode_id = ent.replace(
-                                "http://www.wikidata.org/entity/", ""
-                            )
-                        else:
-                            qnode_id = ent
-
-                        if ":" in qnode_id:
-                            qnode_id, qnode_prob = qnode_id.split(":", 1)
-                            assert fastnumbers.isfloat(qnode_prob)
-                            qnode_prob = fastnumbers.float(qnode_prob)
-                        else:
-                            qnode_prob = 1.0
-                        pents.append((qnode_id, qnode_prob))
-
-                    if has_only_correct_entity:
-                        candidates = [CandidateEntity(pents[0][0], pents[0][1])]
-                    else:
-                        candidates = [CandidateEntity(e[0], e[1]) for e in pents[1:]]
-
-                    link = Link(
-                        start=0,
-                        end=len(table.columns[ci][ri]),
-                        url=f"http://www.wikidata.org/entity/{pents[0][0]}",
-                        entity_id=pents[0][0] if pents[0][0].strip() != "" else None,
-                        candidates=candidates,
-                    )
-                links[ri][ci].append(link)
         return LinkedTable(table, Context(), links)
 
+    @staticmethod
+    def parse_link_file(
+        table: ColumnBasedTable, infile: Union[Path, str], top_k: int = 100
+    ) -> List[List[List[Link]]]:
+        """
+        Each row of a link file has the following format: `<row_index>\t<col_index>\t(<link>|(<entity_id>(\t<entity_id>)*))`, where:
+            * `row_index` and `col_index` start from 0
+            * `row_index` does not count the header of the table (i.e., skip the first row of `infile` if it's the header)
+            * `(<link>|(<entity_id>(\t<entity_id>)*))` is either:
+                - `<link>` a json string encoding Link object and can be deserialized using `Link.from_dict` function
+                - or (<entity_id>(\t<entity_id>)*) a list of entity ids joined by `\t` tab character, each entity id can be a wikidata qnode id (e.g., Q414) or a full qnode uri (e.g., "http://www.wikidata.org/entity/Q414"). The first entity is considered as the correct entity of the cell, and the rest are considered as the candidate entities of the cell
+        Note that a pair `<row_index>`, `<col_index>` don't have to be unique.
+        """
+        links = []
+        nrows, ncols = table.shape()
+
+        for ri in range(nrows):
+            links.append([[] for _ci in range(ncols)])
+
+        rows = []
+        for row in M.deserialize_csv(infile, delimiter="\t"):
+            ri, ci, ents = int(row[0]), int(row[1]), row[2:]
+            rows.append((ri, ci, ents))
+
+        has_only_correct_entity = all(len(ents) == 1 for _, _, ents in rows)
+
+        for ri, ci, ents in rows:
+            if len(ents) == 1 and ents[0].startswith("{"):
+                link = Link.from_dict(orjson.loads(ents[0]))
+            else:
+                pents = []
+                for ent in ents[: top_k + 1]:
+                    if ent.startswith("http"):
+                        assert ent.startswith("http://www.wikidata.org/entity/")
+                        qnode_id = ent.replace("http://www.wikidata.org/entity/", "")
+                    else:
+                        qnode_id = ent
+
+                    if ":" in qnode_id:
+                        qnode_id, qnode_prob = qnode_id.split(":", 1)
+                        assert fastnumbers.isfloat(qnode_prob)
+                        qnode_prob = fastnumbers.float(qnode_prob)
+                    else:
+                        qnode_prob = 1.0
+                    pents.append((qnode_id, qnode_prob))
+
+                if has_only_correct_entity:
+                    candidates = [CandidateEntity(pents[0][0], pents[0][1])]
+                else:
+                    candidates = [CandidateEntity(e[0], e[1]) for e in pents[1:]]
+
+                link = Link(
+                    start=0,
+                    end=len(table.columns[ci][ri]),
+                    url=f"http://www.wikidata.org/entity/{pents[0][0]}",
+                    entity_id=pents[0][0] if pents[0][0].strip() != "" else None,
+                    candidates=candidates,
+                )
+            links[ri][ci].append(link)
+
+        return links
+
 
 @dataclass
 class Context:
@@ -200,6 +233,28 @@ class Context:
     page_url: Optional[str] = None
     page_entity_id: Optional[str] = None
 
+    content_hierarchy: List[ContentHierarchy] = field(default_factory=list)
+
+    def to_dict(self):
+        return {
+            "page_title": self.page_title,
+            "page_url": self.page_url,
+            "page_entity_id": self.page_entity_id,
+            "content_hierarchy": [c.to_dict() for c in self.content_hierarchy],
+        }
+
+    @staticmethod
+    def from_dict(odict: dict):
+        return Context(
+            page_title=odict.get("page_title"),
+            page_url=odict.get("page_url"),
+            page_entity_id=odict.get("page_entity_id"),
+            content_hierarchy=[
+                ContentHierarchy.from_dict(c)
+                for c in odict.get("content_hierarchy", [])
+            ],
+        )
+
 
 @dataclass
 class CandidateEntity:
@@ -210,13 +265,13 @@ class CandidateEntity:
 @dataclass
 class Link:
     start: int
-    end: int
-    url: str
+    end: int  # exclusive
+    url: Optional[str]  # none when there is no link, and the entity is not mapped yet
     entity_id: Optional[str]
     candidates: List[CandidateEntity]
 
     @staticmethod
-    def from_dict(obj: dict) -> "Link":
+    def from_dict(obj: dict) -> Link:
         return Link(
             start=obj["start"],
             end=obj["end"],
@@ -225,9 +280,21 @@ class Link:
             candidates=[CandidateEntity(**c) for c in obj.get("candidates", [])],
         )
 
+    def to_dict(self) -> dict:
+        return {
+            "start": self.start,
+            "end": self.end,
+            "url": self.url,
+            "entity_id": self.entity_id,
+            "candidates": [asdict(c) for c in self.candidates],
+        }
+
     def remove_nonexisting_entities(self, nonexisting_entities: Set[str]):
         if self.entity_id in nonexisting_entities:
             self.entity_id = None
         self.candidates = [
             c for c in self.candidates if c.entity_id not in nonexisting_entities
         ]
+
+    def length(self) -> int:
+        return self.end - self.start
